#!/usr/bin/env python
# coding: utf8

"""
    `spleeter.utils.audio` package provides various
    tools for manipulating audio content such as :

    - Audio adapter class for abstract interaction with audio file.
    - FFMPEG implementation for audio adapter.
    - Waveform convertion and transforming functions.
"""

__email__ = 'spleeter@deezer.com'
__author__ = 'Deezer Research'
__license__ = 'MIT License'
